(in-package :asdf-world)

;;; Pathname utilities

(defun relative-pathname-p (pathname)
  (eq :relative (first (pathname-directory pathname))))

(defun relativize-pathname (pathname reference-pathname)
  "Return a pathname that's equivalent to PATHNAME but relative to
REFERENCE-PATHNAME if possible. Like ENOUGH-NAMESTRING but inserts
..'s if necessary."
  (let ((pathname (merge-pathnames pathname *default-pathname-defaults*))
        (reference-pathname (merge-pathnames reference-pathname
                                             *default-pathname-defaults*)))
    (assert (equal (pathname-host pathname)
                   (pathname-host reference-pathname)))
    (assert (equal (pathname-device pathname)
                   (pathname-device reference-pathname)))
    (let* ((dir (pathname-directory pathname))
           (ref-dir (pathname-directory reference-pathname))
           (mismatch-index (or (mismatch dir ref-dir :test #'equal)
                               (length dir))))
      (make-pathname :directory (nconc (list :relative)
                                       (make-list (- (length ref-dir)
                                                     mismatch-index)
                                                  :initial-element :up)
                                       (subseq dir mismatch-index))
                     :defaults pathname))))

(defun pathname->relname (pathname &optional reference)
  "Return a portable representation of the relative PATHNAME."
  (when reference
    (setq pathname (relativize-pathname pathname reference)))
  (unless (relative-pathname-p pathname)
    (error "~S is not relative." pathname))
  (list (rest (pathname-directory pathname))
        (pathname-name pathname)
        (pathname-type pathname)))

(defun relname->pathname (relname &optional reference)
  "Inverse of PATHNAME->RELNAME."
  (let ((pathname (make-pathname :directory (cons :relative (first relname))
                                 :name (second relname)
                                 :type (third relname))))
    (if reference
        (merge-pathnames pathname reference)
        pathname)))

;;; WORLD

(defvar *world* nil
  "The file in which the current world is stored or NIL.")

(defclass world ()
  ((pathname
    :initarg :pathname
    :reader world-pathname
    :documentation "Truename of the file in which this world is stored.")
   (install-location
    :initarg :install-location
    :reader %install-location
    :documentation "Pathname to where ASDF-INSTALL may install
packages. Setfable.")
   (system-map
    :initform nil
    :initarg :system-map
    :accessor system-map
    :documentation "A list of list that maps system names to portable
pathnames relative to PATHNAME.")))

(defun install-location (&optional (world *world*))
  (let ((world (world world)))
    (%install-location world)))

(defun load-world (pathname)
  "Load the world object stored in pathname."
  (let ((pathname (truename pathname)))
    (destructuring-bind (&key install-location system-map)
        (with-open-file (s pathname) (read s))
      (make-instance 'world
                     :pathname pathname
                     :install-location (merge-pathnames
                                        (relname->pathname install-location)
                                        (make-pathname :name nil :type nil
                                                       :defaults pathname))
                     :system-map system-map))))

(defun save-world (world)
  "Save WORLD to where it was loaded from."
  (with-standard-io-syntax
    (with-open-file (s (world-pathname world)
                     :direction :output :if-does-not-exist :create
                     :if-exists :supersede)
      ;; Let's try to be somewhat version control friendly and emit
      ;; one line per entry.
      (format s "(~S~% ~S~% ~S~% (~{~S~^~%  ~}))"
              :install-location (pathname->relname (install-location world)
                                                   (world-pathname world))
              :system-map (sort (system-map world) #'string<
                                :key #'first))))
  world)

(defun create-world (pathname &key (install-location (make-pathname)))
  "Create an empty world with INSTALL-LOCATION and save it to
PATHNAME. If INSTALL-LOCATION is relative it is relative to PATHNAME."
  (save-world (make-instance
               'world :pathname pathname
               :install-location (merge-pathnames
                                  install-location
                                  (make-pathname :name nil :type nil
                                                 :defaults pathname)))))

(defun world (world-designator)
  "Return the WORLD identified by WORLD-DESIGNATOR that is either a
world object or a pathname."
  (if (typep world-designator 'world)
      world-designator
      (load-world world-designator)))

(defun world-system-pathname (system &optional (world *world*))
  "Return the pathname of the asdf SYSTEM in this WORLD or NIL."
  (let* ((world (world world))
         (worldpath (second (find (asdf::coerce-name system) (system-map world)
                                  :key #'first :test #'string=))))
    (if worldpath
        (relname->pathname worldpath (world-pathname world))
        nil)))

(defun set-world-system-pathname (pathname system world)
  "Set the location of the asdf file for SYSTEM to PATHNAME in this WORLD."
  (let* ((world (world world))
         (name (asdf::coerce-name system))
         (entry (find name (system-map world) :key #'first :test #'string=))
         (worldpath (pathname->relname pathname (world-pathname world))))
    (if entry
        (setf (second entry) worldpath)
        (setf (system-map world)
              (cons (list name worldpath) (system-map world))))
    (save-world world)
    pathname))

(defsetf world-system-pathname (system &optional (world '*world*)) (store)
  `(set-world-system-pathname ,store ,system ,world))

(defun remove-system (system &optional (world *world*))
  "Remove SYSTEM from WORLD."
  (let* ((world (world world))
         (name (asdf::coerce-name system))
         (entry (find name (system-map world) :key #'first :test #'string=)))
    (when entry
      (setf (system-map world)
            (remove name (system-map world) :key #'first :test #'string=))
      (save-world world)
      t)))

(defun set-install-location (pathname &optional (world *world*))
  "Set WORLD's INSTALL-LOCATION to PATHNAME. If it is relative it is
relative to the file that holds WORLD."
  (let ((world (world world)))
    (setf (slot-value world 'install-location)
          (merge-pathnames pathname
                           (make-pathname :name nil :type nil
                                          :defaults (world-pathname world))))
    (save-world world)
    pathname))

(defsetf install-location set-install-location)

(defun map-world (function &optional (world *world*))
  "Map FUNCTION over the canonical system names in WORLD."
  (let ((world (world world)))
    (map nil (lambda (entry)
               (funcall function (first entry)))
         (system-map world))))

;;; Plumbing for ASDF and ASDF-INSTALL

(defun search-world (system)
  (when (and *world* (probe-file *world*))
    (world-system-pathname system)))

(pushnew 'search-world asdf:*system-definition-search-functions*)

;;; Add a location that always refers to the current world.
(let ((location-name "Current world"))
  (unless (find location-name asdf-install:*locations*
                :key #'third :test #'equal)
    (setf asdf-install:*locations*
          ;; These expressions are eval'ed by our modified asdf-install:
          (cons (list '(when *world* (install-location (world *world*)))
                      '*world*
                      location-name)
                asdf-install:*locations*))))
